#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <gst/gst.h>

void play_file(const gchar *filepath, const gchar *confStr)
{
	GstStateChangeReturn sret;
	GstElement *pipeline;
	GstElement *src;
	GstMessage *msg = NULL;
	GError *error = NULL;
	GstBus *bus;

	/* initialise the gstreamer system */
	gst_init(NULL, NULL);

	g_print("Trying to play stream...\n");

	/* create pipeline */
	pipeline = gst_parse_launch("rspsrc name=rsp_source ! queue2 max-size-bytes=32768 use-buffering=TRUE ! decodebin ! autoaudiosink", &error);
	if(!pipeline) {
		g_print ("Pipeline parse error: %s\n", error->message);
		return;
	}
	/* get source element from pipeline, by name */
	src = gst_bin_get_by_name(GST_BIN (pipeline), "rsp_source");
	/* set rsp stream config file, or config-json, etc */
	if(filepath)	
		g_object_set(src, "config-file", filepath, NULL);
	if(confStr)
		g_object_set(src, "config-json", confStr, NULL);
	g_object_set(src, "send-status", 1, "send-raw", 1, NULL); 

	/* get the pipeline's bus, for receiving messages, tags, etc. */
	bus = gst_element_get_bus(pipeline);
	
	/* and start the pipeline running */
	gst_element_set_state(GST_ELEMENT(pipeline), GST_STATE_PLAYING);
	
	/* Blocks until state change either completes or fails */
	sret = gst_element_get_state(GST_ELEMENT(pipeline), NULL, NULL, -1);
	switch(sret){
		case GST_STATE_CHANGE_FAILURE:{
			msg = gst_bus_poll(bus, GST_MESSAGE_ERROR, 0);
			if(msg){
				GError *err = NULL;
				gchar *dbg_str = NULL;
				
				gst_message_parse_error (msg, &err, &dbg_str);
				g_printerr("FAILED to play: %s\n%s\n", err->message,
							(dbg_str) ? dbg_str : "(no debugging information)");
				g_error_free(err);
				g_free (dbg_str);
				gst_message_unref (msg);
			}else
				g_printerr ("FAILED to play: unknown error\n");
			break;
		}
		case GST_STATE_CHANGE_SUCCESS:{
			GstMessage *msg;
			
			g_print("Playing ...\n");
			
			GstStructure *stats_struct;
			unsigned int cluster_cnt;
			float fec;
			float err;
			float bad;
			float dup;
			float buffer;
			
			while(1){	
				msg = gst_bus_poll(bus, GST_MESSAGE_TAG | GST_MESSAGE_EOS | GST_MESSAGE_ERROR, 0);
				if(msg && GST_MESSAGE_TYPE(msg) == GST_MESSAGE_ERROR){
					GError *err = NULL;
					gchar *dbg_str = NULL;

					gst_message_parse_error (msg, &err, &dbg_str);
					g_printerr("ERROR: %s\n%s\n", err->message,
							   (dbg_str) ? dbg_str : "(no debugging information)");
					g_error_free(err);
					g_free (dbg_str);
					gst_message_unref(msg);
					continue;
				}
				
				if(msg && GST_MESSAGE_TYPE(msg) == GST_MESSAGE_EOS){
					g_print ("Finished.\n");
					break;
				}
				
				if(msg && GST_MESSAGE_TYPE(msg) == GST_MESSAGE_TAG){
					GstTagList *tags = NULL;
					gchar *tag_str = NULL;

					gst_message_parse_tag (msg, &tags);
					g_print("Tags from element %s:\n", GST_OBJECT_NAME(msg->src));
					if(tag_str = gst_tag_list_to_string(tags)){
						g_print("%s\n", tag_str);
						g_free(tag_str);
					}
					gst_tag_list_unref (tags);
					continue;
				}
				
				/* sleep for one second */
				g_usleep (G_USEC_PER_SEC * 1);
				g_object_get(src, "stats", &stats_struct, NULL);
				if(stats_struct){
					if(gst_structure_get(stats_struct, 
						    "cluster_size", G_TYPE_UINT, &cluster_cnt,
						    "buffer_%", G_TYPE_FLOAT, &buffer,
						    "correction_%", G_TYPE_FLOAT, &fec,
						    "err_pkt_%", G_TYPE_FLOAT, &err,
						    "dup_pkt_%", G_TYPE_FLOAT, &dup,
						    "bad_pkt_%", G_TYPE_FLOAT, &bad, NULL))
					  g_print("\33[2K\rRSP: Cluster=%d Buffer=%.0f Fixed=%.0f Err=%.0f Bad=%.0f Dup=%.0f", 
								cluster_cnt, buffer, fec, err, bad, dup);
				}
			}
			break;
		}
		default:
			g_assert_not_reached();
	}
	
	/* shut down and free everything */
	gst_element_set_state(pipeline, GST_STATE_NULL);
	gst_object_unref(pipeline);
	gst_object_unref(bus);
	return;
}

int main(int argc, char *argv[])
{
	if(argc == 2)
		play_file(argv[1], NULL);
	else if(argc == 1){
		size_t count, total;
		char *text = (char *)malloc(257);
		total = 0;
		while(1){
			count = fread(text+total, 1, 256, stdin);
			total = total + count;
			if(count == 256)
				text = realloc(text, total + 257);
			else
				break;
		}
		text[total] = 0;
		play_file(NULL, text);
	}else{
		fprintf(stderr, "Please provide a single command argument: a file path to the rsp listen file,\n");
		fprintf(stderr, "or no command argument to read a rsp configuration jSON string from stdin.\n");
	}
	return 0;
}
